dofile 'sgd_ccs.lua'

local nid

local pold = parameters * 0.0
local pserver = parameters * 0.0 -- parameters at the server
local pservero = parameters * 0.0 -- previous epoch parameters at the server
local nupdateso = parameters * 0.0 -- previous epoch statistics about parameters' update at the server
local delta = parameters * 0.0
local tempdelta = parameters * 0.0
local nup = math.ceil(opt.uploadFraction * parameters:nElement())
local ndown = math.ceil(opt.downloadFraction * parameters:nElement())
local inx, inx_, accurcy, cmptm

if paths.filep(resultsfile) then
   results = torch.load(resultsfile, 'binary')
   print('load results ' .. resultsfile)
else 
   results = {}
end

if paths.filep(nupdatesfile) then
   nupdates = torch.load(nupdatesfile, 'binary')
else
   nupdates = parameters * 0.0 -- statistics about parameters' update at the server
end

pserver:copy(parameters)

pservero:copy(parameters)

local p = {} -- parameters of all nodes
local g = {} -- gradeints of all nodes

for nid = 1, opt.netSize do
   g[nid] = parameters * 0.0
end

if paths.filep(pfile) then
   p = torch.load(pfile, 'binary')
else
   for nid = 1, opt.netSize do
      p[nid] = parameters * 0.0
      p[nid]:copy(parameters)
   end
end

parameters:copy(pserver)

for i = 1,opt.nepochs do
   if epoch >= (opt.nepochs * opt.netSize) then
      break
   end

   testnid = torch.random(opt.netSize)

   nupdateso:copy(nupdates) -- previous epoch statistics
   pservero:copy(pserver)

   nupdates:mul(0.8) -- fade out the old update statistics

   for nid = 1, opt.netSize do
      if opt.method ~= 'seq' then
         nid = torch.random(opt.netSize)
      end

      -- get the latest parameters of node nid
      parameters:copy(p[nid])

      if opt.method == 'asy' and torch.rand(1)[1] > 0.5 then
         getoldpar = 1
      else
         getoldpar = 0
      end
      -- download a fraction of server's parameters
      if opt.downloadFraction < 1 and opt.downloadFraction > 0 then
         if getoldpar == 1 then
            _,inx = nupdateso:sort(1)
           for elmnt = 1, ndown do
              inx_ = inx[{ elmnt - ndown - 1 }]
              parameters[inx_] = pservero[inx_]
           end
         else
            _,inx = nupdates:sort(1)
           for elmnt = 1, ndown do
              inx_ = inx[{ elmnt - ndown - 1 }]
              parameters[inx_] = pserver[inx_]
           end
         end
         ---
      elseif opt.downloadFraction == 1 then
         if getoldpar == 1 then
            parameters:copy(pservero)
         else
            parameters:copy(pserver)
         end
      end

      pold:copy(parameters)

      accurcy = train(nid)

      local r = {}
      r[1] = epoch
      r[2] = nid
      r[3] = accurcy
      r[4] = 'train'
      table.insert(results, r);

      print('train ' .. nid .. ' ' .. accurcy .. ' ' .. epoch)

      -- don't test all the time - to speed up the process
      if nid == testnid then
         accurcy = test()

         local r = {}
         r[1] = epoch
         r[2] = nid
         r[3] = accurcy
         r[4] = 'test'
         table.insert(results, r);

         print('test ' .. nid .. ' ' .. accurcy .. ' ' .. epoch)
      end

      -- save the parameters and gradient of node nid
      p[nid]:copy(parameters)
      g[nid]:copy(parameters - pold)

      if opt.range > 0 then
         -- impose the bound, and determine the sensitivity
         g[nid]:apply(function(x) return cut(x, opt.range) end)
      end

      -- privacy: sparse vector
      if opt.range > 0 and opt.epsilon > 0 then
         eps = opt.epsilon / (2 * opt.range)
         if opt.dpver == 'book' then
            if opt.delta > 0 then
               sqrt512 = math.sqrt(512)
               eps_T = (sqrt512 / (sqrt512 + 1)) * eps
               eps_R = (2 / (sqrt512 + 1)) * eps
               std_T = math.sqrt(32 * nup * math.log(2 / opt.delta)) / eps_T
               std_R = math.sqrt(32 * nup * math.log(2 / opt.delta)) / eps_R
            else
               eps_T = (8/9) * eps
               eps_R = (2/9) * eps
               std_T = (2 * nup) / eps_T
               std_R = (2 * nup) / eps_R
            end
            std_C = 2 * std_T
         elseif opt.dpver == 'notes' then
            -- ignores opt.delta 
            std_T = 2 / eps
	    std_R = (2 * nup) / eps
            std_C = std_T
         end
      end

      delta:copy(g[nid])

      thrshld = opt.threshold
      local maxgrad = 0

      -- Just moify the threshold so that we run our original algorithm
      if opt.epsilon == 0 and opt.threshold == 0 then
         tempdelta:copy(delta)
         _,inx = tempdelta:abs():sort(1)
         if nup > 0 then
            thrshld = tempdelta[inx[-nup]]
         else
            thrshld = 1 + tempdelta[inx[-1]]
         end
         maxgrad = tempdelta[inx[-1]]
      end
      --

      local perm = torch.randperm(parameters:nElement())
      local count = 0

      if opt.range > 0 and opt.epsilon > 0 then
         local noiseT = lapnoise(std_T, 1)
         thrshld = opt.threshold + noiseT[1]
      end

      for elmnt = 1, parameters:nElement() do
         local eid = perm[elmnt]
         upval = 0
          
         if opt.range > 0 and opt.epsilon > 0 then
            local noiseC = lapnoise(std_C, 2)

            if (delta[eid] + noiseC[1]) >= thrshld or (-delta[eid] + noiseC[2]) >= thrshld then
               local noiseR = lapnoise(std_R, 1)
               upval = (delta[eid] + noiseR[1]) 

               local noiseT = lapnoise(std_T, 1)
               thrshld = opt.threshold + noiseT[1]
            end
         elseif opt.epsilon == 0 then
            if math.abs(delta[eid]) >= thrshld then
               upval = delta[eid]
            end
         end

         if upval ~= 0 then
            if opt.range > 0 then
               upval = cut(upval, opt.range)
            end
            pserver[eid] = pserver[eid] + upval
            nupdates[eid] = nupdates[eid] + 1
            count = count + 1
         end

      
         if count >= nup then
            break
         end
      end
         
      -- print('release', maxgrad, thrshld, count, nup, count / math.max(nup,1)) -- what fraction of what we wanted to upload we actually uploaded

   end

   -- serialize
   torch.save(epochfile, epoch, 'binary')
   torch.save(paramfile, parameters, 'binary')
   torch.save(pfile, p, 'binary')
   torch.save(nupdatesfile, nupdates, 'binary')
   torch.save(resultsfile, results, 'binary')
end

-- test the server
parameters:copy(pserver)
accurcy = test()
local r = {}
r[1] = epoch
r[2] = 0
r[3] = accurcy
r[4] = 'test'
table.insert(results, r);
print('test ' .. 0 .. ' ' .. accurcy .. ' ' .. epoch)

-- test each of the participants
local avrg = 0

for nid = 1, opt.netSize do
   parameters:copy(p[nid])
   accurcy = test()

   local r = {}
   r[1] = epoch
   r[2] = '-1' -- last test
   r[3] = accurcy
   r[4] = 'test'
   table.insert(results, r);

   print('test a' .. nid .. ' ' .. accurcy .. ' ' .. epoch)
   avrg = avrg + accurcy
end

avrg = avrg / opt.netSize

local r = {}
r[1] = epoch + 1
r[2] = 'a' -- average
r[3] = avrg
r[4] = 'test'
table.insert(results, r);

print('test a ' .. avrg .. ' ' .. epoch)

torch.save(resultsfile, results, 'binary')

